# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""VIIRS SDR Satpy reader.

This VIIRS reader is designed for reading the NPP/JPSS SDR HDF5 files.
The input files are produced by CSPP Polar (CSPP RDR pipeline),
and the read by hdf5.

V2.0.0:  NRL-Monterey, 082025

"""

# Python Standard Libraries
import logging
import os
import re
from itertools import chain
from datetime import datetime, timedelta

# Third-Party Libraries
import h5py
import numpy as np
import xarray as xr
from scipy.interpolate import NearestNDInterpolator

# geoips libraries
from geoips.data_manipulations.corrections import apply_gamma

LOG = logging.getLogger(__name__)

try:
    from lunarref.lib.liblunarref import lunarref

    lunarref_lib = True
except ImportError:
    lunarref_lib = False
    LOG.warning("Failed lunarref import, if needed, install it.")


interface = "readers"
family = "standard"
name = "viirs_sdr_hdf5"
source_names = ["viirs"]


VARLIST = {
    "DNB": ["DNB"],
    "IMG": ["I04", "I05"],
    "IMG-Vis": ["I01", "I02", "I03"],
    "MOD": ["M07", "M08", "M10", "M11", "M12", "M13", "M14", "M15", "M16"],
    "MOD-Vis": ["M01", "M02", "M03", "M04", "M05", "M06", "M09"],
}


def bowtie_correction(band, lat, lon):
    """Correct input data for the instrument bowtie effect.

    Correction derived from: Remote Sens. 2016, 8, 79.
    """
    # Unfold
    ord_lat = np.sort(lat, axis=0)
    unfold_idx = np.argsort(lat, axis=0)

    rad_fold = np.take_along_axis(band, unfold_idx, axis=0)
    sort_lon = np.take_along_axis(lon, unfold_idx, axis=0)

    if np.all(np.isnan(rad_fold)):
        LOG.debug("All nan band, no bowtie correction")
        return rad_fold, ord_lat, sort_lon
    if np.all(np.isnan(rad_fold)):
        LOG.debug("All NaNs, no bowtie correction")
        return rad_fold, ord_lat, sort_lon

    res_band = rad_fold.copy()
    # mask and indicies of bad data
    rad_nan_flag = np.isnan(rad_fold)
    nan_mask = np.where(~rad_nan_flag)

    interp = NearestNDInterpolator(np.transpose(nan_mask), rad_fold[nan_mask])
    interp_nans = interp(*np.indices(rad_fold.shape)[:, rad_nan_flag], **{"workers": 4})
    res_band[rad_nan_flag] = interp_nans

    return res_band, ord_lat.astype(np.float64), sort_lon.astype(np.float64)


def get_time(hfile):
    """Get datetime from the metadata."""
    with h5py.File(hfile) as h5data:
        base_key = list(h5data["Data_Products"].keys())[0]
        main_key = [
            i for i in h5data["Data_Products"][base_key].keys() if "Gran_0" in i
        ][0]

        start_time = h5data["Data_Products"][base_key][main_key].attrs[
            "N_Beginning_Time_IET"
        ][0][0]
        end_time = h5data["Data_Products"][base_key][main_key].attrs[
            "N_Ending_Time_IET"
        ][0][0]

    stime = timedelta(microseconds=int(start_time)) + datetime.strptime(
        "1958-01-01", "%Y-%m-%d"
    )
    etime = timedelta(microseconds=int(end_time)) + datetime.strptime(
        "1958-01-01", "%Y-%m-%d"
    )

    return stime, etime


def get_plat(hfile):
    """Get platform from the metadata."""
    with h5py.File(hfile) as h5data:
        plat = h5data.attrs["Platform_Short_Name"][0][0].astype("str")
    platform_dict = {
        "NPP": "Suomi-NPP",
        "JPSS-1": "NOAA-20",
        "J01": "NOAA-20",
        "JPSS-2": "NOAA-21",
        "J02": "NOAA-21",
    }
    plat_name = platform_dict[plat]
    return plat_name


def get_sci_data(hfile, sv_key):
    """Get science data from file."""
    with h5py.File(hfile) as h5data:
        band_key = list(h5data["All_Data"].keys())[0]
        # get BT and Radiance
        bt_flag = False
        ref_flag = False

        if "BrightnessTemperature" in h5data["All_Data"][band_key].keys():
            bt_raw = h5data["All_Data"][band_key]["BrightnessTemperature"][...]
            if "BrightnessTemperatureFactors" in h5data["All_Data"][band_key].keys():
                bt_fac = h5data["All_Data"][band_key]["BrightnessTemperatureFactors"][
                    ...
                ]
                bt_cal = bt_raw * bt_fac[0] + bt_fac[1]
            else:
                bt_cal = bt_raw
            bt_flag = True
        elif "Reflectance" in h5data["All_Data"][band_key].keys():

            bt_raw = h5data["All_Data"][band_key]["Reflectance"][...]
            if "ReflectanceFactors" in h5data["All_Data"][band_key].keys():
                bt_fac = h5data["All_Data"][band_key]["ReflectanceFactors"][...]
                bt_cal = bt_raw * bt_fac[0] + bt_fac[1]
            else:
                bt_cal = bt_raw
            bt_cal = apply_gamma(bt_cal, 1.65) * 100
            ref_flag = True

        rad_raw = h5data["All_Data"][band_key]["Radiance"][...]
        if "RadianceFactors" in h5data["All_Data"][band_key].keys():
            rad_fac = h5data["All_Data"][band_key]["RadianceFactors"][...]
            LOG.info("Applying calibration values to radiance data.")
            rad_cal = rad_raw * rad_fac[0] + rad_fac[1]
        else:
            rad_cal = rad_raw
        # flag data
        flag_key = [k for k in h5data["All_Data"][band_key].keys() if "QF1_" in k][0]
        qflag = h5data["All_Data"][band_key][flag_key][...]

    mask = np.where(qflag > 1, True, False)
    if not np.isdtype(rad_cal.dtype, np.float32):
        rad_cal = rad_cal.astype(np.float32)
    # seems like DNB quality flags are bad??
    # leads to NaN granules
    rad_cal[mask] = np.nan

    if bt_flag or ref_flag:
        if not np.isdtype(bt_cal.dtype, np.float32):
            bt_cal = bt_cal.astype(np.float32)
        bt_cal[mask] = np.nan

    sci_xr = xr.Dataset()
    if bt_flag:
        sci_xr[sv_key + "BT"] = xr.DataArray(bt_cal, dims=("dim_0", "dim_1"))
    if ref_flag:
        sci_xr[sv_key + "Ref"] = xr.DataArray(bt_cal, dims=("dim_0", "dim_1"))

    sci_xr[sv_key + "Rad"] = xr.DataArray(rad_cal, dims=("dim_0", "dim_1"))
    sci_xr["Quality_Flag"] = xr.DataArray(qflag, dims=("dim_0", "dim_1"))

    return sci_xr


def get_geo_data(bname, band_type, flist):
    """Get geolocation data based upon the band provided.

    Input
    -----
    bname: basename of input file
    band_type: band type (DNB, SVI, SVM)
    flist: file list

    Output:
    -------
    dict: output parameters
    latitude:
    longitude:
    """
    BAD_VAL = -998.0
    geo_dict = {"SVI": "GITCO", "SVD": "GDNBO", "SVM": "GMTCO"}
    geo_key = geo_dict[band_type]

    # get proper geo file and at the right time
    set_start = re.search(r"_t\d{7}_", bname)[0]
    flist = [f for f in flist if set_start in f]

    base_flist = list(map(os.path.basename, flist))
    geo_list = [flist[i] for i, j in enumerate(base_flist) if geo_key in j]
    if len(geo_list) == 0:
        # try to find ellispoid files
        LOG.warning("No GEO terrain corrected files found, reverting to ellispoid.")
        geo_dict = {"SVI": "GIMGO", "SVM": "GMODO"}
        # catch error for DNB files
        if band_type not in geo_dict.keys():
            raise LookupError("No VIIRS geo-file found, check input filelist.")
        geo_key = geo_dict[band_type]
        geo_list = [flist[i] for i, j in enumerate(base_flist) if geo_key in j]

    if len(geo_list) == 0:
        raise LookupError("No VIIRS GEO file found, check input filelist.")
    geo_file = geo_list[0]
    LOG.info("Reading geo file {}".format(os.path.basename(geo_file)))
    with h5py.File(geo_file) as h5geo:
        geo_key = list(h5geo["All_Data"].keys())[0]
        latitude = h5geo["All_Data"][geo_key]["Latitude"][...]
        longitude = h5geo["All_Data"][geo_key]["Longitude"][...]
        mask = h5geo["All_Data"][geo_key]["QF2_VIIRSSDRGEO"][...]

        sol_zea = h5geo["All_Data"][geo_key]["SolarZenithAngle"][...]
        sol_azi = h5geo["All_Data"][geo_key]["SolarAzimuthAngle"][...]

        sat_zea = h5geo["All_Data"][geo_key]["SatelliteZenithAngle"][...]
        sat_azi = h5geo["All_Data"][geo_key]["SatelliteAzimuthAngle"][...]
        extra_params = {}
        if "DNB" in geo_key:
            latitude = (
                h5geo["All_Data"][geo_key]["Latitude_TC"]
                if "latitude_tc" in h5geo["All_Data"][geo_key].keys()
                else h5geo["All_Data"][geo_key]["Latitude"][...]
            )
            longitude = (
                h5geo["All_Data"][geo_key]["Longitude_TC"]
                if "longitude_tc" in h5geo["All_Data"][geo_key].keys()
                else h5geo["All_Data"][geo_key]["Longitude"][...]
            )

            lza = h5geo["All_Data"][geo_key]["LunarZenithAngle"][...]
            laa = h5geo["All_Data"][geo_key]["LunarAzimuthAngle"][...]
            moon_frac = h5geo["All_Data"][geo_key]["MoonIllumFraction"][...]
            moon_phase_ang = h5geo["All_Data"][geo_key]["MoonPhaseAngle"][...]

            # moon_frac = 0.5 * (1 + np.cos(moon_phase_ang))
            lza = np.where(lza < BAD_VAL, np.nan, lza)
            laa = np.where(laa < BAD_VAL, np.nan, laa)

            extra_params = {
                "lza": lza,
                "laa": laa,
                "moon_fraction": moon_frac,
                "moon_phase_angle": moon_phase_ang,
            }

    # mask and flagging
    flag = np.where(mask > 0, True, False)
    latitude[flag] = np.nan
    longitude[flag] = np.nan

    if np.any(latitude < -90) or np.any(latitude > 90):
        lat_flag = np.where(latitude < -90, True, False)
        lat_flag = np.logical_or(np.where(latitude > 90, True, False), lat_flag)
        latitude[lat_flag] = np.nan

    comb_mask = (
        (sol_zea < BAD_VAL)
        | (sol_azi < BAD_VAL)
        | (sat_zea < BAD_VAL)
        | (sat_azi < BAD_VAL)
        | flag
    )
    sol_zea = np.where(sol_zea < BAD_VAL, np.nan, sol_zea)
    sol_azi = np.where(sol_azi < BAD_VAL, np.nan, sol_azi)

    sat_zea = np.where(sat_zea < BAD_VAL, np.nan, sat_zea)
    sat_azi = np.where(sat_azi < BAD_VAL, np.nan, sat_azi)
    # create xarray dataset
    geo_xr = xr.Dataset()
    geo_xr["latitude"] = xr.DataArray(latitude, dims=("dim_0", "dim_1"))
    geo_xr["longitude"] = xr.DataArray(longitude, dims=("dim_0", "dim_1"))
    geo_xr["solar_zenith_angle"] = xr.DataArray(sol_zea, dims=("dim_0", "dim_1"))
    geo_xr["solar_azimuth_angle"] = xr.DataArray(sol_azi, dims=("dim_0", "dim_1"))
    geo_xr["satellite_zenith_angle"] = xr.DataArray(sat_zea, dims=("dim_0", "dim_1"))
    geo_xr["satellite_azimuth_angle"] = xr.DataArray(sat_azi, dims=("dim_0", "dim_1"))
    geo_xr["geo_mask"] = xr.DataArray(comb_mask, dims=("dim_0", "dim_1"))

    if len(extra_params.keys()) > 0:
        geo_xr["lunar_zenith_angle"] = xr.DataArray(
            extra_params["lza"], dims=("dim_0", "dim_1")
        )
        geo_xr["lunar_azimuth_angle"] = xr.DataArray(
            extra_params["laa"], dims=("dim_0", "dim_1")
        )
        geo_xr["moon_frac"] = xr.DataArray(
            np.full(sol_zea.shape, extra_params["moon_fraction"] / 100.0),
            dims=("dim_0", "dim_1"),
        )

        geo_xr["moon_phase"] = xr.DataArray(
            np.full(sol_zea.shape, extra_params["moon_phase_angle"]),
            dims=("dim_0", "dim_1"),
        )

    return geo_xr


def call(
    fnames,
    metadata_only=False,
    chans=None,
    area_def=None,
    self_register=False,
    bowtie=True,
):
    """Read VIIRS SDR hdf5 data products.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * Return before actually reading data if True
    chans : list of str, default=None
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * NOT YET IMPLEMENTED
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * NOT YET IMPLEMENTED
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.
    bowtie: bool, default=True
        * Preform bowtie correction for edge of scan.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.
        * Conforms to geoips xarray standards, see more in geoips documentation.
    """
    base_fnames = list(map(os.path.basename, fnames))
    sorted_fnames = sorted(fnames)

    dt_start = get_time(sorted_fnames[0])[0]
    dt_end = get_time(sorted_fnames[-1])[-1]
    plat_name = get_plat(sorted_fnames[0])
    meta_attrs = {
        "source_file_name": base_fnames,
        "start_datetime": dt_start,
        "end_datetime": dt_end,
        "source_name": "viirs",
        "platform_name": plat_name,
        "data_provider": "NOAA",
        "sample_distance_km": 1,
        "interpolation_radius_of_influence": 2000,
    }

    attr_xr = xr.Dataset(attrs=meta_attrs)
    attr_dict = {"METADATA": attr_xr}
    if metadata_only:
        return attr_dict

    full_xr = {}
    # trim VARLIST based on channels requested
    if chans:
        tmp_vl = VARLIST.copy()
        for key, val in tmp_vl.items():
            km = [c[:3] in val for c in chans]
            if not any(km):
                VARLIST.pop(key)
            else:
                matches = [s for s in val for c in chans if s in c]
                VARLIST[key] = matches

    # create list from requested keys
    vlist = list(chain.from_iterable(VARLIST.values()))
    trim_flist = [f for f in fnames for v in vlist if "SV" + v in os.path.basename(f)]
    # parse all science files
    for scifile in trim_flist:

        fn_base = os.path.basename(scifile)
        band_fname = re.search(r"SV\w{1}(\d{2}|\w{2})", fn_base)[0]
        LOG.info("Parsing VIIRS file; band {}, filename {}".format(band_fname, fn_base))
        band_type = band_fname[:3]
        band_spec = band_fname[2:]

        sci_xrt = get_sci_data(scifile, band_spec)

        # apply children xr to parents
        parent_key = [k for k, v in VARLIST.items() if band_spec in v][0]
        # update, append, or create xarray
        if parent_key in full_xr.keys():
            if band_spec + "Rad" in list(full_xr[parent_key].variables):
                # append
                geo_xrt = get_geo_data(fn_base, band_type, fnames)
                sci_xrt.update(geo_xrt)
                full_xr[parent_key] = xr.concat(
                    [full_xr[parent_key], sci_xrt], dim="dim_0"
                )
            else:
                # same geo, new data, same band
                full_xr[parent_key].update(sci_xrt)
        else:
            # new geo, new data, new band
            full_xr[parent_key] = sci_xrt
            geo_xrt = get_geo_data(fn_base, band_type, fnames)
            full_xr[parent_key].update(geo_xrt)
            full_xr[parent_key].attrs = meta_attrs
    if bowtie:
        # update data
        for k, dset in full_xr.items():
            tmp_lat = dset["latitude"].data
            tmp_lon = dset["longitude"].data
            # trim_dset = dset.drop_vars(["latitude","longitude"])
            var_avail = list(dset.variables)

            rad_list = [i for i in var_avail if "Rad" in i]
            rad_map = {}
            for rad_name in rad_list:
                rad_var = dset[rad_name].data
                rad_bow, slat, slon = bowtie_correction(rad_var, tmp_lat, tmp_lon)
                rad_map[rad_name] = xr.DataArray(rad_bow, dims=("dim_0", "dim_1"))
            bt_list = [i for i in var_avail if "Ref" in i or "BT" in i]
            bt_map = {}
            for bt_name in bt_list:
                bt_var = dset[bt_name].data
                bt_bow, slat, slon = bowtie_correction(bt_var, tmp_lat, tmp_lon)
                bt_map[bt_name] = xr.DataArray(bt_bow, dims=("dim_0", "dim_1"))

            udict = {
                "latitude": xr.DataArray(slat, dims=("dim_0", "dim_1")),
                "longitude": xr.DataArray(slon, dims=("dim_0", "dim_1")),
            }
            udict.update(bt_map)
            udict.update(rad_map)
            full_xr[k].update(udict)

    if "DNBRef" in chans or chans is None:
        lunar_ref_data = lunarref(
            full_xr["DNB"]["DNBRad"].data,
            full_xr["DNB"]["solar_zenith_angle"].data,
            full_xr["DNB"]["lunar_zenith_angle"].data,
            meta_attrs["start_datetime"].strftime("%Y%m%d%H"),
            meta_attrs["end_datetime"].strftime("%M"),
            full_xr["DNB"]["moon_phase"].data,
        )
        full_xr["DNB"]["DNBRef"] = xr.DataArray(lunar_ref_data, dims=("dim_0", "dim_1"))

    full_xr |= attr_dict

    return full_xr



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Pytest file for calling integration bash scripts."""

import os
import shlex
from datetime import datetime, timezone
import re

import pytest

from geoips.geoips_utils import call_cmd
from geoips.filenames.base_paths import PATHS as gpaths

print(
    "\nOUTPUT_CHECKER_THRESHOLD_IMAGE: "
    f"{os.environ.get('OUTPUT_CHECKER_THRESHOLD_IMAGE')}\n"
)

# Quick test to ensure basic functionality in geoips installation.
# This test list should NOT require/test anything in any other plugin package.
base_integ_test_calls = [
    "$geoips_repopath/tests/scripts/amsr2.config_based_no_compare.sh",
    "$geoips_repopath/tests/scripts/amsr2_ocean.tc.windspeed.imagery_clean.sh",
]

# Linting integration tests, ensure code and documentation are correctly formatted.
lint_integ_test_calls = [
    "$geoips_repopath/tests/utils/check_code.sh all $geoips_repopath",
    "$geoips_repopath/docs/build_docs.sh $geoips_repopath $geoips_pkgname html_only",
]

# This includes ALL test scripts within the geoips repo itself with available data,
# excluding the tests under base and lint markers.
# This test list should NOT require/test anything in any other plugin package.
full_integ_test_calls = [
    "$geoips_repopath/tests/scripts/abi.static.Infrared.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/abi.static.Infrared.imagery_annotated_enhanced.sh",
    "$geoips_repopath/tests/scripts/console_script_create_sector_image.sh",
    "$geoips_repopath/tests/scripts/abi.static.dmw.imagery_windbarbs_high.sh",
    "$geoips_repopath/tests/scripts/abi.static.Test-Visible-Logarithmic.imagery_clean.sh",  # noqa: E501
    "$geoips_repopath/tests/scripts/abi.static.Visible.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/abi.static.nasa_dust_rgb.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/abi.config_based_dmw_overlay.sh",
    "$geoips_repopath/tests/scripts/abi.config_based_output_full_zarr_cache_backend.sh",
    "$geoips_repopath/tests/scripts/abi.config_based_output_low_memory.sh",
    "$geoips_repopath/tests/scripts/abi.config_based_output.sh",
    "$geoips_repopath/tests/scripts/ami.static.Infrared.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/ami.static.Visible.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/ami.static.mst.absdiff-IR-BD.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/ami.tc.WV.geotiff.sh",
    "$geoips_repopath/tests/scripts/ami.WV-Upper.unprojected_image.sh",
    "$geoips_repopath/tests/scripts/amsr2.global.89H-Physical.cogeotiff.sh",
    "$geoips_repopath/tests/scripts/amsr2.tc.89H-Physical.imagery_annotated.sh",
    "$geoips_repopath/tests/scripts/amsr2_rss.tc.windspeed.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/amsr2.config_based_overlay_output.sh",
    "$geoips_repopath/tests/scripts/amsr2.config_based_overlay_output_low_memory.sh",
    "$geoips_repopath/tests/scripts/amsua_mhs_mirs.tc.rainrate.imagery.sh",
    "$geoips_repopath/tests/scripts/ascat_knmi.tc.windbarbs.imagery_windbarbs_clean.sh",
    "$geoips_repopath/tests/scripts/ascat_low_knmi.tc.windbarbs.imagery_windbarbs.sh",
    "$geoips_repopath/tests/scripts/ascat_noaa_25km.tc.windbarbs.imagery_windbarbs.sh",
    (
        "$geoips_repopath/tests/scripts/"
        "ascat_noaa_50km.tc.wind-ambiguities.imagery_windbarbs.sh"
    ),
    "$geoips_repopath/tests/scripts/ascat_uhr.tc.wind-ambiguities.imagery_windbarbs.sh",
    "$geoips_repopath/tests/scripts/ascat_uhr.tc.nrcs.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/ascat_uhr.tc.windbarbs.imagery_windbarbs.sh",
    "$geoips_repopath/tests/scripts/ascat_uhr.tc.windspeed.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/atms.tc.165H.netcdf_geoips.sh",
    "$geoips_repopath/tests/scripts/aws_TB50.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/aws_TB89.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/aws_TB165.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/aws_TB180.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/aws_TB325-1.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/aws_tc_TB50.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/aws_tc_TB89.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/aws_tc_TB165.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/aws_tc_TB180.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/aws_tc_TB325-1.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/cli_dummy_script.sh",
    "$geoips_repopath/tests/scripts/fci.static.Visible.imagery_annotated.sh",
    "$geoips_repopath/tests/scripts/fci.unprojected_image.Infrared.sh",
    "$geoips_repopath/tests/scripts/gfs.static.windspeed.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/gfs.static.waveheight.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/gmi.tc.89pct.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/imerg.tc.Rain.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/mhs_mirs.tc.183-3H.imagery_annotated.sh",
    "$geoips_repopath/tests/scripts/mimic_coarse.static.TPW-CIMSS.imagery_annotated.sh",
    "$geoips_repopath/tests/scripts/mimic_fine.tc.TPW-PWAT.imagery_annotated.sh",
    "$geoips_repopath/tests/scripts/modis.Infrared.unprojected_image.sh",
    "$geoips_repopath/tests/scripts/oscat_knmi.tc.windbarbs.imagery_windbarbs.sh",
    "$geoips_repopath/tests/scripts/sar.tc.nrcs.imagery_annotated.sh",
    "$geoips_repopath/tests/scripts/seviri.WV-Upper.unprojected_image.sh",
    "$geoips_repopath/tests/scripts/seviri.airmass.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/seviri.Convective_Storms.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/seviri.Day_Microphys_Summer.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/seviri.Day_Microphys_Winter.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/seviri.Day_Solar.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/seviri.Dust_RGB.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/seviri.Natural_Color.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/seviri.Night_Microphys.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/seviri.Volcanic_Ash.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/sgli.static.IR-RGB.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/smap.unsectored.text_winds.sh",
    "$geoips_repopath/tests/scripts/smos.tc.sectored.text_winds.sh",
    "$geoips_repopath/tests/scripts/viirs.static.visible.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/viirsday.global.Night-Vis-IR.cogeotiff_rgba.sh",
    "$geoips_repopath/tests/scripts/viirsday.tc.Night-Vis-IR.imagery_annotated.sh",
    "$geoips_repopath/tests/scripts/viirsmoon.tc.Night-Vis-GeoIPS1.imagery_clean.sh",
    (
        "$geoips_repopath/tests/scripts/"
        "seviri.WV-Upper.no_self_register.unprojected_image.sh"
    ),
    (
        "$geoips_repopath/tests/scripts/"
        "viirsclearnight.Night-Vis-IR-GeoIPS1.unprojected_image.sh"
    ),
]

# Test scripts spanning multiple repositories / geoips plugins.
multi_repo_integ_test_calls = [
    "python $GEOIPS_PACKAGES_DIR/geoips/tests/utils/test_interfaces.py",
    "$geoips_repopath/tests/scripts/console_script_list_available_plugins.sh",
]

# Test scripts that require test datasets with limited availability.
limited_data_integ_test_calls = [
    "$geoips_repopath/tests/scripts/ewsg.static.Infrared.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/hy2.tc.windspeed.imagery_annotated.sh",
    "$geoips_repopath/tests/scripts/saphir.tc.183-3HNearest.imagery_annotated.sh",
    "$geoips_repopath/tests/scripts/ssmi.tc.37pct.imagery_clean.sh",
    "$geoips_repopath/tests/scripts/ssmis.color91.unprojected_image.sh",
]


def set_log_filename(expanded_call):
    """Set the log filename for individual pytest outputs."""
    script_name = expanded_call[0]
    # Strip out any full paths, if a test has a full path passed in
    # Also replace any "*" with "-"
    args_str = ""
    for curr_arg in expanded_call[1:]:
        if "/" in curr_arg:
            curr_arg = os.path.basename(curr_arg)
        args_str = f"{args_str}_{curr_arg}".replace("*", "-")
    curr_datetime = datetime.utcnow().strftime("%Y%m%d.%H%M%S")
    basename_str = os.path.basename(script_name)
    log_fname = f"{os.environ['GEOIPS_OUTDIRS']}/logs/pytests/integration"
    log_fname = f"{log_fname}/{curr_datetime}.{basename_str}{args_str}.log"
    return log_fname


def check_environment(check_environment_script):
    """Call check environment script to confirm environment is set appropriately."""
    print("Checking environment")
    cmd = ["bash", check_environment_script]
    log_fname = set_log_filename([os.path.basename(check_environment_script)])
    print(f"Check environment log: {log_fname}")
    retval, stdout, stderr = call_cmd(
        cmd, output_log_fname=log_fname, use_logging=False, use_print=False
    )
    if retval != 0:
        print(f"\nCall to '{check_environment_script}' failed, exiting")
        print("See error output in log:")
        print(f"Log: {log_fname}")
        raise RuntimeError(
            f"Check environment failed, please see error output in log {log_fname}\n"
        )


def check_base_install():
    """
    Run the base installation check script to verify the GeoIPS installation.

    Executes the 'base_install.sh' script located in the GeoIPS tests directory
    to ensure that all required components are properly installed.

    Raises
    ------
    subprocess.CalledProcessError
        If the installation check script returns a non-zero exit status.
    """
    cmd = [
        "bash",
        os.path.join(
            os.getenv("GEOIPS_PACKAGES_DIR"),
            "geoips/tests/integration_tests/base_install.sh",
        ),
        "exit_on_missing",
        "skip_create_registries",
    ]
    # print("Running base install check")
    script_name = "base_install.sh"
    log_fname = set_log_filename([script_name])
    retval, stdout, stderr = call_cmd(
        cmd, output_log_fname=log_fname, use_logging=False, use_print=False
    )
    if retval != 0:
        print(f"\nCall to '{script_name}' failed, exiting")
        print("See error output in log:")
        print(f"Log: {log_fname}")
        raise RuntimeError(
            f"{script_name} failed, please see error output in log {log_fname}\n"
        )


def check_full_install():
    """
    Run the full installation check script to verify the GeoIPS installation.

    Executes the 'full_install.sh' script located in the GeoIPS tests directory
    to ensure that all required components are properly installed.

    Raises
    ------
    subprocess.CalledProcessError
        If the installation check script returns a non-zero exit status.
    """
    cmd = [
        "bash",
        os.path.join(
            os.getenv("GEOIPS_PACKAGES_DIR"),
            "geoips/tests/integration_tests/full_install.sh",
        ),
        "exit_on_missing",
        "skip_create_registries",
    ]
    # print("Running full install check")
    script_name = "full_install.sh"
    log_fname = set_log_filename([script_name])
    retval, stdout, stderr = call_cmd(
        cmd, output_log_fname=log_fname, use_logging=False, use_print=False
    )
    if retval != 0:
        print(f"\nCall to '{script_name}' failed, exiting")
        print("See error output in log:")
        print(f"Log: {log_fname}")
        raise RuntimeError(
            f"{script_name} failed, please see error output in log {log_fname}\n"
        )


def check_site_install():
    """
    Run the site installation check script to verify the GeoIPS installation.

    Executes the 'site_install.sh' script located in the GeoIPS tests directory
    to ensure that all required components are properly installed.

    Raises
    ------
    subprocess.CalledProcessError
        If the installation check script returns a non-zero exit status.
    """
    cmd = [
        "bash",
        os.path.join(
            os.getenv("GEOIPS_PACKAGES_DIR"),
            "geoips/tests/integration_tests/site_install.sh",
        ),
        "exit_on_missing",
        "skip_create_registries",
    ]
    # print("Running site install check")
    script_name = "site_install.sh"
    log_fname = set_log_filename([script_name])
    retval, stdout, stderr = call_cmd(
        cmd, output_log_fname=log_fname, use_logging=False, use_print=False
    )
    if retval != 0:
        print(f"\nCall to '{script_name}' failed, exiting")
        print("See error output in log:")
        print(f"Log: {log_fname}")
        raise RuntimeError(
            f"{script_name} failed, please see error output in log {log_fname}\n"
        )


def setup_environment():
    """
    Set up necessary environment variables for integration tests.

    Configures paths and package names for the GeoIPS core and its plugins by
    setting environment variables required for the integration tests. Assumes
    that 'GEOIPS_PACKAGES_DIR' is already set in the environment.

    Notes
    -----
    The following environment variables are set:
    - geoips_repopath
    - geoips_pkgname
    """
    # Base path
    os.environ["geoips_repopath"] = os.path.realpath(
        os.path.join(os.path.dirname(__file__), "..", "..")
    )
    os.environ["geoips_pkgname"] = "geoips"

    # Environment variable for GEOIPS_PACKAGES_DIR
    geoips_packages_dir = os.getenv("GEOIPS_PACKAGES_DIR")
    if not geoips_packages_dir:
        raise EnvironmentError("GEOIPS_PACKAGES_DIR environment variable not set.")


@pytest.fixture(scope="session")
def site_setup():
    """
    Set up the site integration tests by checking install and setting env-vars.

    Calls `check_site_install()` to verify the installation and `setup_environment()`
    to set up the necessary environment variables before running integration tests.
    """
    check_site_install()


@pytest.fixture(scope="session")
def base_setup():
    """
    Set up the base integration tests by checking install and setting env-vars.

    Calls `check_base_install()` to verify the installation and `setup_environment()`
    to set up the necessary environment variables before running integration tests.
    """
    check_base_install()


@pytest.fixture(scope="session")
def full_setup():
    """
    Set up the full integration tests by checking install and setting env-vars.

    Calls `check_full_install()` to verify the installation and `setup_environment()`
    to set up the necessary environment variables before running integration tests.
    """
    check_full_install()


def is_likely_oserror_missing_file(log_contents: str) -> bool:
    """
    Check if the log contents indicate a likely OSError for a missing file.

    Searches the last ten lines of the provided log contents for patterns
    that suggest an OSError related to a missing file.

    Parameters
    ----------
    log_contents : str
        The contents of the log file to check.

    Returns
    -------
    bool
        True if the log contents indicate a likely OSError for a missing file,
        False otherwise.
    """
    last_ten_lines = "\n".join(log_contents.strip().splitlines()[-10:])
    pattern = r"OSError: (File '[^']+' not found|No files found on disk)\."
    return re.search(pattern, last_ten_lines) is not None


def run_script_with_bash(script, fail_on_missing_data, unset_output_path_env_vars=True):
    """
    Run scripts by executing specified shell commands with bash.

    If unset_output_path_env_vars is set, explicit output path env vars are unset to
    ensure clean consistent paths for test comparisons (these env vars all default
    to a consistent location if not set, so ensure they are not set so output products
    are written to a consistent location for tests).


    Parameters
    ----------
    script : str
        Shell command to executes. The command may
        contain environment variables which will be expanded before execution.

    Raises
    ------
    subprocess.CalledProcessError
        If the shell command returns a non-zero exit status.
    """
    # The print statements all appear to print AFTER the script is complete
    # And appears to not include a newline after printing the script.
    print("")
    # Unset all supported specific output path env vars to ensure consistent test
    # output locations.  These are all defaulted to GEOIPS_OUTDIRS locations in
    # geoips_utils replace paths - we want to ensure they are not set so all paths
    # are updated with the GEOIPS_OUTDIRS relative paths for test consistency across
    # installs. GEOIPS_REPLACE_OUTPUT_PATHS is set within geoips/filenames/base_paths.py
    # to a default set of supported environment variables, overridden to the
    # specified list if GEOIPS_REPLACE_OUTPUT_PATHS is set within the environment.
    # See geoips/geoips/filenames/base_paths.py for more information on
    # GEOIPS_REPLACE_OUTPUT_PATHS
    removed_env_vars = {}
    if unset_output_path_env_vars:
        for env_var in gpaths["GEOIPS_REPLACE_OUTPUT_PATHS"]:
            if env_var in os.environ:
                removed_env_vars[env_var] = os.environ[env_var]
                del os.environ[env_var]
        if removed_env_vars:
            print(f"REMOVED ENV VARS {removed_env_vars.keys()}")
    else:
        print("NOT REMOVING ENV VARS!")

    if ".sh" in script:
        expanded_call = shlex.split("bash " + os.path.expandvars(script))
    else:
        expanded_call = shlex.split(os.path.expandvars(script))
    # print("Running: ", script)
    # print("Expanded call: ", expanded_call)
    # print(" ".join(expanded_call))
    log_fname = set_log_filename(expanded_call[1:])

    # Note - this never seems to print until after the cmd is complete
    print(f"Log: {log_fname} , latest logs:")
    print(f"ls -lthr {os.path.dirname(log_fname)}/*")
    print(datetime.now(timezone.utc))
    retval, stdout, stderr = call_cmd(
        expanded_call,
        output_log_fname=log_fname,
        use_logging=False,
        use_print=False,
        pipe=True,
    )
    if removed_env_vars:
        print(f"RESTORING ENV VARS {removed_env_vars.keys()}")
        for env_var in removed_env_vars:
            os.environ[env_var] = removed_env_vars[env_var]
    if retval != 0:
        with open(log_fname, "r") as logfile:
            log_contents = logfile.read()
            print("\nLOG FILE CONTENTS:\n")
            print(log_contents)
            print("-" * 80)
        print("FAILED COMMAND FOR PREVIOUS TEST")
        print(" ".join(expanded_call))
        print(f"FAILED LOG FILE: {log_fname}")
        print("FAILED")

        if is_likely_oserror_missing_file(log_contents):
            if fail_on_missing_data:
                raise FileNotFoundError(
                    f"FileNotFoundError, see output in log {log_fname}\n"
                )
            else:
                pytest.xfail(f"FileNotFoundError, see output in log {log_fname}\n")
        else:
            raise RuntimeError(f"CalledProcessError, see output in log {log_fname}\n")
    else:
        print("PASSED COMMAND FOR PREVIOUS TEST")
        print(" ".join(expanded_call))
        print(f"PASSED LOG FILE: {log_fname}")
        print("PASSED")


@pytest.mark.base
@pytest.mark.integration
@pytest.mark.parametrize("script", base_integ_test_calls)
def test_integ_base_test_script(
    base_setup: None, script: str, fail_on_missing_data: bool
):
    """
    Run integration test scripts by executing specified shell commands.

    Parameters
    ----------
    script : str
        Shell command to execute as part of the integration test. The command may
        contain environment variables which will be expanded before execution.

    Raises
    ------
    subprocess.CalledProcessError
        If the shell command returns a non-zero exit status.
    """
    setup_environment()
    run_script_with_bash(script, fail_on_missing_data)


@pytest.mark.lint
@pytest.mark.integration
@pytest.mark.parametrize("script", lint_integ_test_calls)
def test_integ_lint_test_script(base_setup: None, script: str):
    """
    Run integration test scripts by executing specified shell commands.

    Parameters
    ----------
    script : str
        Shell command to execute as part of the integration test. The command may
        contain environment variables which will be expanded before execution.

    Raises
    ------
    subprocess.CalledProcessError
        If the shell command returns a non-zero exit status.
    """
    setup_environment()
    run_script_with_bash(script, fail_on_missing_data=True)


@pytest.mark.full
@pytest.mark.spans_multiple_packages
@pytest.mark.integration
@pytest.mark.parametrize("script", multi_repo_integ_test_calls)
def test_integ_multi_repo_script(
    site_setup: None, script: str, fail_on_missing_data: bool
):
    """
    Run integration test scripts by executing specified shell commands.

    Parameters
    ----------
    script : str
        Shell command to execute as part of the integration test. The command may
        contain environment variables which will be expanded before execution.

    Raises
    ------
    subprocess.CalledProcessError
        If the shell command returns a non-zero exit status.
    """
    setup_environment()
    run_script_with_bash(script, fail_on_missing_data)


@pytest.mark.full
@pytest.mark.integration
@pytest.mark.parametrize("script", full_integ_test_calls)
def test_integ_full_script(full_setup: None, script: str, fail_on_missing_data: bool):
    """
    Run integration test scripts by executing specified shell commands.

    Parameters
    ----------
    script : str
        Shell command to execute as part of the integration test. The command may
        contain environment variables which will be expanded before execution.

    Raises
    ------
    subprocess.CalledProcessError
        If the shell command returns a non-zero exit status.
    """
    setup_environment()
    run_script_with_bash(script, fail_on_missing_data)


# These are required to test the full functionality of this repo, but have limited
# test dataset availability.
@pytest.mark.full
@pytest.mark.optional
@pytest.mark.integration
@pytest.mark.parametrize("script", limited_data_integ_test_calls)
def test_integ_limited_data_script(
    full_setup: None, script: str, fail_on_missing_data: bool
):
    """
    Run integration test scripts by executing specified shell commands.

    Parameters
    ----------
    script : str
        Shell command to execute as part of the integration test. The command may
        contain environment variables which will be expanded before execution.

    Raises
    ------
    subprocess.CalledProcessError
        If the shell command returns a non-zero exit status.
    """
    setup_environment()
    run_script_with_bash(script, fail_on_missing_data)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Data algorithm for High Near constant contrast imagery product.

Algorithm is based upon work by Zinke (2017) for improved NCC VIIRS DNB imagery.
"""
# python libs
import logging

# installed libs
import numpy as np

# geoips libs
from geoips.data_manipulations.corrections import apply_data_range

LOG = logging.getLogger(__name__)

interface = "algorithms"
family = "list_numpy_to_numpy"
name = "high_near_constant_contrast"


def gain_func(angle):
    """Gain factor for angle correction.

    Angle in degrees.
    """
    # given an ndarray
    # create masks and then calculate based upon the masks
    intervals = [-np.inf, 87.541, 96, 101, 103.49, np.inf]

    masks = []
    for i, j in enumerate(intervals[:-1]):
        tmp_mask = np.logical_and(angle > j, angle <= intervals[i + 1])
        masks.append(tmp_mask)

    gain_fac = np.full(angle.shape, np.nan)

    # angle <= 87.541:
    gain_fac[masks[0]] = (58 + (4 / np.cos(angle[masks[0]]))) / 5
    # angle <= 96.00 and angle > 87.541:
    gain_fac[masks[1]] = (123 * np.exp(1.06 * (angle[masks[1]] - 89.589))) * (
        (((angle[masks[1]] - 93) ** 2) / 18) + 0.5
    )
    # angle <= 101 and angle > 96:
    gain_fac[masks[2]] = 123 * np.exp(1.06 * (angle[masks[2]] - 89.589))
    # angle <= 103.49 and angle > 101:
    gain_fac[masks[3]] = (123 * np.exp(1.06 * (101 - 89.589))) * np.log(
        angle[masks[3]] - (101 - np.e)
    ) ** 2
    # else:
    gain_fac[masks[4]] = 6e7

    return gain_fac


def call(
    arrays,
):
    """Calculate HNCC.

    Must have solar/lunar zenith angle and moon illumination fraction.
    """
    # check data
    if len(arrays) != 4:
        raise ValueError(
            "Incongruent shape, please pass data,solar_zenith_angle,lunar_zenith_angle"
        )

    data, sza, lza, moon_ill_frac = arrays

    gain_solar = gain_func(sza)
    gain_lza = gain_func(lza)
    LOG.info(
        "Producing HNCC imagery with moon illumination of {0:.2f}".format(
            np.nanmean(moon_ill_frac)
        )
    )
    # calculate the lunar brightness
    # apparent solar magnitude
    mag_sol = -26.74
    # mean apparent mag of moon
    mam_moon_mag = -12.74

    mag_moon_variation = (
        0.026 * np.arccos(2 * moon_ill_frac - 1)
        + (np.arccos(2 * moon_ill_frac - 1) ** 4) * 4e-9
    )
    mag_moon = mag_moon_variation + mam_moon_mag
    bt_ratio = 10 ** ((mag_sol - mag_moon) / -2.5)
    gain_lunar = bt_ratio * gain_lza
    # systematic bias for radiances
    # from Liao et al. 2013, 12712
    bias_rad = data + 2.6e-10

    total_gain = 1 / ((1 / gain_solar) + (1 / gain_lunar))
    norm_rad = bias_rad * total_gain
    hncc_arr = apply_data_range(
        norm_rad,
        min_val=0.0,
        max_val=0.2,
        min_outbounds="mask",
        max_outbounds="mask",
        norm=True,
        inverse=False,
    )
    return hncc_arr
